import re

def fix_paragraph(x):
    # replace what look like hyphenations
    x = re.sub(r"(\S+)-[\n\r]\s*(\S+)", r"\1\2", x)

    # replace newlines with spaces
    x = re.sub(r"[\n\r]", " ", x)

    # replace multiple spaces
    x = re.sub(r" +", " ", x)

    return x

def follow_expand( from_what, follow_what ):
    """
    A SpaCy dependency graph follower...
    :param from_what:
    :param follow_what:
    :return:
    """
    found = [x for x in from_what.children if x.dep_ == follow_what]
    return [
        " ".join(str(x) for x in one.subtree)
        for one in found
    ]

def untokenize(tokens):
    import string
    return "".join([" " + i if not i.startswith("'") and i not in string.punctuation else i for i in tokens]).strip()

def getname(x):
    x = x.split("/")[-1]
    x = re.sub(r'(\.xml|\.txt)', '', x)
    return x



# The following methods are used for extracting in-text citations from the body of a document...

import re
import calendar

# note that 6 is a common mistake for vowels!?
name = r"[A-Z][A-Za-z6'`-̈]+"
# we allow only a single break in the name
name = name + r"\s?" + r"[A-Za-z6'`-̈]*"
fname = r"(?:{name}(?:\s{name})?|{name} et al.?)".format(**locals())
names = r"(?P<names>{fname}(?:, {fname})*(?:,? (and|&) {fname})?)".format(**locals())
year = r"(?P<year>(((\[[0-9]{4}\]\s?)?[0-9]{4})[abcdef]?(, )?)+)"
ppnum = r"pp?\.? (?P<pnum>[0-9\-–—]+)"
pnum = r"(?P<pnum>[0-9\-–—]+)"
prefix = r"(?:(?P<prefix>((see|see also|listed in|and|e.g.|cf.|c.f.),?\s?)+) )?"
postfix = r"(?:(?P<postfix>,?.*))?"

regex_in_order = [
    r"^{prefix}{names}, {year}{postfix}$".format(**locals()),
    r"^{prefix}{names} {year}, {ppnum}{postfix}$".format(**locals()),
    r"^{prefix}{names} {year}{postfix}$".format(**locals()),
    r"^{prefix}{names}, {year}: {pnum}{postfix}$".format(**locals()),
    r"^{prefix}{names} \[{year}\]{postfix}$".format(**locals()),
    r"^{prefix}{names} \[{year}, [^\]]\]{postfix}$".format(**locals()),
]

# print(regex_in_order)

name_blacklist = calendar.month_name[1:]

faulty_years = set()


def getOuterParens(x):
    paren_level = 0
    buffer = ""
    rets = []
    i_start = None

    for i, c in enumerate(x):
        if c == "(":
            if paren_level == 0:
                i_start = i
            paren_level += 1

            if paren_level == 2:  # hack to remove multiple levels
                paren_level = 1

        if c == ")":
            paren_level -= 1

            if paren_level <= 0:
                rets.append((i_start, buffer[1:]))

                paren_level = 0  # hack to remove multiple levels

                buffer = ""
        if paren_level > 0:
            buffer += c

    return rets


def extractCitation(t):
    t = re.sub(r"[-]\n", "", t)
    t = re.sub(r"\s+", " ", t)

    # try splitting it both ways,
    # just see which gets more hits

    tparts = t.split(";")
    # print("A")
    ret = []
    for tpart in tparts:
        tpart = tpart.strip()
        # print("Aa")
        # I guess this causes infinite looping??
        if len(tpart.split(",")) >= 4:
            continue
        # print("Ab")
        # print("Checking", tpart)
        for rgxi, rgx in enumerate(regex_in_order):
            mtch = re.match(rgx, tpart)
            if mtch:
                d = mtch.groupdict()
                if d['names'] in name_blacklist:
                    continue
                ret.append(d)
                break  # wow... didn't have this before and was getting annoying duplicates smh...
    # print("B")
    tparts2 = t.split(",")

    ret2 = []
    for tpart in tparts2:
        tpart = tpart.strip()

        # print("Checking", tpart)
        for rgx in regex_in_order:
            mtch = re.match(rgx, tpart)
            if mtch:
                d = mtch.groupdict()
                if d['names'] in name_blacklist:
                    continue
                ret2.append(d)
    # print("C")
    if len(ret2) > len(ret):
        return ret2
    return ret


def extractCitations(strs):
    results = []
    for s in strs:
        result = extractCitation(s)
        if result is not None:
            results.append(result)
    return results


# demos coming from https://guides.dataverse.org/en/latest/api/native-api.html#add-file-api

# NOTE just found PyDataverse: https://pydataverse.readthedocs.io/en/latest/_modules/pyDataverse/utils.html
# easy package for doing this stuff (so I don't have to fuck it up so much...)

from . import env
import requests
from pathlib import Path


def confirm(pmt = "",):
    """
    Ask user to enter Y or N (case-insensitive).
    :return: True if the answer is Y.
    :rtype: bool
    """
    answer = ""
    while answer not in ["y", "n"]:
        answer = input(f"{pmt} [Y/N]? ").lower()
    return answer == "y"


def download(
    persistent_id,
    dataverse_server = 'https://dataverse.harvard.edu',
    api_key = None
):

    dv_url = f"{dataverse_server}/api/datasets/:persistentId/?persistentId={persistent_id}"
    if api_key is not None:
        dv_url += f"&key={api_key}"

    r = requests.get(dv_url)
    res = r.json()

    if res['status'] != 'OK':
        raise Exception("There was an issue retrieving the dataset. Please check your persistent ID, and provide an API key if this is a DRAFT version.")

    v = res['data']['latestVersion']
    title = [x for x in v['metadataBlocks']['citation']['fields'] if x['typeName'] == 'title'][0]['value']

    print(f"Retrieving latest version of f{title}. datasetId = {v['datasetId']}. last updated {v['lastUpdateTime']}.")
    print(f"{len(v['files'])} files. Total file size = {sum(x['dataFile']['filesize'] for x in v['files']) / 1000000:0.1f} Mb")
    
    outFold = env.variable_dir.joinpath(title)

    while outFold.exists():
        print(f"The target folder {outFold} exists. Prompting to overwrite.")
        if confirm("Overwrite?"):
            break

        outFoldName = input(f"Enter a new folder to store dataset {title}")
        outFold = env.variable_dir.joinpath(outFoldName)

    if outFold.exists():
        raise Exception("I'm too scared to overwrite. Aborting")

    print("Downloading zip file...")

    down = f"{dataverse_server}/api/access/dataset/:persistentId/?persistentId={persistent_id}"
    headers = {'X-Dataverse-key':api_key}
    r = requests.get(down, headers=headers, allow_redirects=True)

    tmp_file = env.variable_dir.joinpath('tmp.zip')
    tmp_file.open('wb').write( r.content )

    print(f"Done! Unzipping into {outFold.name}")
    import zipfile
    with zipfile.ZipFile(tmp_file, 'r') as zip_ref:
        zip_ref.extractall(outFold)


def upload( 
    api_key, 
    dataset_name,
    dataset_description="", 
    dataverse_name = 'kk-citation-counts',
    dataverse_server = 'https://dataverse.harvard.edu'):


    to_upload = Path(env.variable_dir).joinpath( dataset_name )
    if not to_upload.exists():
        raise Exception("Please provide a valid `dataset_name`, one which exists. It should be a folder in the `knowknow` variable directory.")
        
    # no trailing slash

    metadata = {
    "datasetVersion": {
        "metadataBlocks": {
        "citation": {
            "fields": [
            {
                "value": dataset_name,
                "typeClass": "primitive",
                "multiple": False,
                "typeName": "title"
            },
            {
                "value": [
                {
                    "authorName": {
                    "value": "McGail, Alec",
                    "typeClass": "primitive",
                    "multiple": False,
                    "typeName": "authorName"
                    },
                    "authorAffiliation": {
                    "value": "Cornell University",
                    "typeClass": "primitive",
                    "multiple": False,
                    "typeName": "authorAffiliation"
                    }
                }
                ],
                "typeClass": "compound",
                "multiple": True,
                "typeName": "author"
            },
            {
                "value": [ 
                    { "datasetContactEmail" : {
                        "typeClass": "primitive",
                        "multiple": False,
                        "typeName": "datasetContactEmail",
                        "value" : "am2873@cornell.edu"
                    },
                    "datasetContactName" : {
                        "typeClass": "primitive",
                        "multiple": False,
                        "typeName": "datasetContactName",
                        "value": "McGail, Alec"
                    }
                }],
                "typeClass": "compound",
                "multiple": True,
                "typeName": "datasetContact"
            },
            {
                "value": [ {
                "dsDescriptionValue":{
                    "value":   f"""\
Count dataset for Web of Science citation data.
{dataset_description}
Produced and uploaded by the Python package <a href='https://pypi.org/project/knowknow-amcgail/'>knowknow</a>""",
                    "multiple":False,
                "typeClass": "primitive",
                "typeName": "dsDescriptionValue"
                }}],
                "typeClass": "compound",
                "multiple": True,
                "typeName": "dsDescription"
            },
            ],
            "displayName": "Citation Metadata"
        }
        }
    }
    }

    import json
    import requests

    # -------------------
    # Start by creating a new dataset
    # -------------------

    url_dataset_id = '%s/api/dataverses/%s/datasets?key=%s' % (dataverse_server, dataverse_name, api_key)

    params_as_json_string = json.dumps(metadata)
    payload = dict(jsonData=params_as_json_string)

    r = requests.post(url_dataset_id, data=params_as_json_string)

    persistentId = r.json()['data']['persistentId']
    dataset_id = r.json()['data']['id']

    print(f"Created dataset with persistent ID = {persistentId}")

    for fn in to_upload.glob("*"):

        # --------------------------------------------------
        # Prepare "file"
        # --------------------------------------------------

        #F_SOURCE = 'D:/knowknow/data/sociology-wos-74b/'
        if 'doc ___ c' in fn.name:
            continue

        print("uploading", fn.name, '...')

        assert fn.exists()

        file_content = fn.open('rb').read()
        files = {'file': (fn.name, file_content)}

        # --------------------------------------------------
        # Using a "jsonData" parameter, add optional description + file tags
        # --------------------------------------------------
        params = dict()
        params_as_json_string = json.dumps(params)
        payload = dict(jsonData=params_as_json_string)

        # --------------------------------------------------
        # Add file using the Dataset's id
        # --------------------------------------------------
        url_dataset_id = '%s/api/datasets/%s/add?key=%s' % (dataverse_server, dataset_id, api_key)

        # -------------------
        # Make the request
        # -------------------
        r = requests.post(url_dataset_id, data=payload, files=files)



from .. import *
from .count_cache import Dataset, make_cross
from ..exceptions import invalid_entry
from . import stats


__all__ = [
    'ttDF', 'TimeTrend'
]

class ttDF:

    def __init__(self, dataset_name, dtype, debug=False):
        self.meta = defaultdict(int)
        self.debug = debug

        self.dataset_name = dataset_name
        self.dataset = Dataset(dataset_name)
        self.dtype = dtype
        self.final = {}

    def run(self):
        self.generate_timetrends()
        self.persist()

    def persist(self):
        varname = "%s.ysum" % (self.dtype)
        self.dataset.save_variable(varname, self.final)
        print('success!')

    def generate_timetrends(self):
        import tracemalloc

        tracemalloc.start()

        items = list(self.dataset.items(self.dtype))

        for ci, item in enumerate(items):

            if ci % 50000 == 0:
                print("Item %s/%s (%s)" % (ci, len(items), item))
                current, peak = tracemalloc.get_traced_memory()
                print(f"Current memory usage is {current / 10**6}MB; Peak was {peak / 10**6}MB")
                print(self.meta)

            if self.debug and ci >= 50000:
                break

            dp = self.make_tt_dump( item )
            if dp is None:
                continue
            self.final[item] = dp

        tracemalloc.stop()

    def make_tt_dump(self, item):
        # this function has an outstanding memory leak!
        self.meta['at least one citation'] += 1

        if self.dataset['type'] == 'wos' and self.dtype == 'c':
            sp = item.split("|")
            if len(sp) < 2:
                self.meta['not enough parts'] += 1
                #raise Exception('wut', item)
                return None

        # memory-inexpensive check...
        if self.dataset(**{self.dtype: item}).docs < 5:
            self.meta['less than 5 citations... dropped.2'] += 1
            return None

        # create a timetrend
        try:
            tt = TimeTrend(name=item, dtype=self.dtype, dataset=self.dataset)
        except invalid_entry:
            self.meta['invalid entry'] += 1
            return None

        # small error catch
        if hasattr(tt, 'pub') and tt.first < tt.pub:
            self.meta['first citation before pub date'] += 1
            return None

        # don't care about those with only a single citation
        if tt.total < 5:
            self.meta['less than 5 citations... dropped.'] += 1
            return None

        # we really don't care about those that never rise in use
        #if tt.first == tt.maxpropy:
        #    self.meta['never rise'] += 1
        #    continue

        self.meta['passed tests pre-blacklist'] += 1

        # set some attributes before it's persisted

        dp = tt.dump()
        return dp




class TimeTrend:
    
    def __init__(self, dataset, dtype, name):
        self.dataset = dataset
        self.dtype = dtype

        self.data_start = self.dataset['RELIABLE_DATA_STARTS_HERE']
        self.data_end = self.dataset['RELIABLE_DATA_ENDS_HERE']
        
        self._cumsum = None
        self._c = None
        self._d = None

        # this marks the end of adding non-saving variables to the class
        self.dict_filter = set(list(self.__dict__) + ["dict_filter"])  # save dict keys

        if type(name) == tuple:
            self.name = name[0]
        else:
            self.name = name

        self.simple_stats()

    def dump(self):
        to_return_keys = set(self.__dict__).difference(self.dict_filter)
        
        return {
            k: getattr(self,k)
            for k in to_return_keys
            if k[0] != '_'
        }
    
    @classmethod
    def addinit(cls, st):
        if st not in cls.init_stack:
            cls.init_stack.append(st)

    def cits(self, A=None, B=None):
        if A is None:
            A = self.data_start
        if B is None:
            B = self.data_end
        return [self.c[YY] for YY in range(A, B+1)]
        
    def docs(self, A=None, B=None):
        if A is None:
            A = self.data_start
        if B is None:
            B = self.data_end
        return [self.d[YY] for YY in range(A, B+1)]

    def show(self):
        plt.plot(
            range(self.data_start, self.data_end+1),
            self.cits(self.data_start, self.data_end),
            label=self.name
        )

    def load_c(self):
        # these counts act as the data for all subsequent computations
        my_by_y = self.dataset.by(self.dtype, 'fy').cits
        all_by_y = self.dataset.by('fy').cits

        self._c = defaultdict(lambda:None)
        self._cp = defaultdict(lambda:None)

        #self._d = defaultdict(lambda:None)
        #self._dp = defaultdict(lambda:None)

        for YY in range( 
            self.data_start,
                self.data_end + 1
        ):
            cy = make_cross(**{self.dtype:self.name, 'fy':YY})
            self._c[YY] = my_by_y[cy]
            self._cp[YY] = my_by_y[cy] / all_by_y[(YY,)] if all_by_y[(YY,)] > 0 else 0

    def load_d(self):
        # these counts act as the data for all subsequent computations
        my_by_y = self.dataset.by(self.dtype, 'fy').docs
        all_by_y = self.dataset.by('fy').docs

        self._d = defaultdict(lambda:None)
        self._dp = defaultdict(lambda:None)

        for YY in range( 
            self.data_start,
                self.data_end + 1
        ):
            dy = make_cross(**{self.dtype:self.name, 'fy':YY})
            self._d[YY] = my_by_y[dy]
            self._dp[YY] = my_by_y[dy] / all_by_y[(YY,)] if all_by_y[(YY,)] > 0 else 0

    @property
    def c(self):
        if self._c is None:
            self.load_c()

        return self._c
    @property
    def cp(self):
        if self._cp is None:
            self.load_c()

        return self._cp
    @property
    def d(self):
        if self._d is None:
            self.load_d()
            
        return self._d
    @property
    def dp(self):
        if self._dp is None:
            self.load_d()
            
        return self._dp
        


    def simple_stats(self):

        non_zero_years = [k for k,c in self.c.items() if c>0]
        if not len(non_zero_years):
            raise invalid_entry('no with positive count', self.name)

        self.first = min(non_zero_years)
        self.last = max(non_zero_years)

        self.maxcounty = max(self.c, key=lambda y:(self.c[y],y))
        self.maxpropy = max(self.cp, key=lambda y:(self.cp[y],y))

        self.maxprop = self.cp[ self.maxpropy ]
        self.maxcount = self.c[ self.maxcounty ]

        self.total = sum(self.c.values())
        self.totalprop = sum(self.cp.values())
        
        if self.dtype == 'c':
            # extracts some extra information from the name
            
            self.type = 'article'
            if self.dataset['type'] == 'wos':
                sp = self.name.split("|")

                if not len(sp):
                    print('Wtf',sp)
                    raise

                try:
                    self.pub = int(sp[1])
                    self.type = 'article'
                except ValueError:
                    self.type = 'book'
                    #self.pub = pubyears[self.name]

            elif self.dataset['type'] == 'jstor':
                inparens = re.findall(r'\(([^)]+)\)', self.name)[0]
                self.pub = int(inparens)

    def avg_between(self, A, B): #not including B
        return self.sum_between(A, B) / (B-A)
            
    def sum_between(self, A, B): #not including B
        if self._cumsum is None:
            # cumsum makes sum_between's very efficient!
            self._cumsum = [self.c[self.data_start]]
            for YY in range(
                    self.data_start + 1,
                    self.data_end + 1
                ):
                self._cumsum.append( self._cumsum[-1] + self.c[YY] )
            self._cumsum = np.array(self._cumsum)



        if A < self.data_start:
            raise Exception("data begins at %s. cannot get %s" % (self.data_start, A))
        if B > self.data_end:
            raise Exception("data ends at %s. cannot get %s" % (self.data_end, B))

        if not (A <= B):
            raise Exception("Can only sum forwards in time. Fix your code...")

        if A == self.data_start:
            return self._cumsum[B - self.data_start]
        else:
            return self._cumsum[B - self.data_start] - self._cumsum[(A - 1) - self.data_start]


    def births_deaths(self, **kwargs):
        """
        def births_deaths(tt,  # the trend this should be calculated on
                          birth_cutoff=0.5,  # relative to the last life
                          death_cutoff=0.1,  # relative to this life so far
                          forward_view=10,  # years to look forward for comparison
                          backward_view=5  # minimum years to look back for comparison
                          ):
        """
        return stats.births_deaths(self, **kwargs)


from .. import defaultdict, np, pd

__all__ = ['top_decade_stratified', 'births_deaths']

def top_decade_stratified(dataset, what, percentile=None, topN=None, yRange=None, maxP=None, debug=True):
    if percentile is None and topN is None:
        print("Assuming top 1%. Use percentile or topN attributes")
        percentile = 0.01

    if percentile is not None and topN is not None:
        raise Exception('one or the other... not both percentile or topN attributes')

    if yRange is None:
        yRange = (
            dataset['RELIABLE_DATA_STARTS_HERE'],
            dataset['RELIABLE_DATA_ENDS_HERE']
        )

    start, end = yRange

    all_tops = set()
    first_top = {}

    # ranges loop from 1940-1950 to 1980-1990, in 1-year increments

    for RANGE_START, RANGE_END in zip(
            range(start, end - 10 + 1),
            range(start + 10, end + 1)
    ):

        count_in_range = defaultdict(int)
        for cross, count in dataset.by(what, 'fy').cits.items():
            if count == 0:
                continue
            if RANGE_END >= cross.fy >= RANGE_START:
                count_in_range[getattr(cross, what)] += count

        counts = list(count_in_range.values())
        if not len(counts):
            if debug: print("Skipping %s" % RANGE_START)
            continue

        if topN is not None and maxP is not None:
            if topN / len(counts) > maxP:
                if debug: print("skipping decade, maxP not reached...")
                continue

        if percentile is not None:
            q99 = np.quantile(np.array(counts), 1 - percentile)
            top1 = {k for k in count_in_range if count_in_range[k] >= q99}
        elif topN is not None:
            top1 = {k for k in sorted(count_in_range, key=lambda x: -count_in_range[x])[:topN]}
        else:
            raise Exception('neverrrr')

        for t in top1:
            if t not in all_tops:
                first_top[t] = RANGE_START
        all_tops.update(top1)

        if percentile is not None:
            if debug: print("%s /%s in the top %0.1f%% in %s,%s (%s total accumulated)" % (
                len(top1),
                len(count_in_range),
                percentile * 100,
                RANGE_START, RANGE_END,
                len(all_tops)
            ))
        else:
            if debug: print("%s /%s  in %s,%s (%s total accumulated)" % (
                len(top1),
                len(count_in_range),
                RANGE_START, RANGE_END,
                len(all_tops)
            ))

    def getrow(name):
        ret = dataset.trend(what, name).dump()
        ret['first_added'] = first_top[name]
        return ret

    alldf = pd.DataFrame.from_records([
        getrow(name)
        for name in all_tops
    ])

    alldf.fillna(value=np.nan, inplace=True)

    print(alldf.shape)
    return alldf


def births_deaths(tt,  # the trend this should be calculated on
                  birth_cutoff=0.5,  # relative to the last life
                  death_cutoff=0.1,  # relative to this life so far
                  forward_view=10,  # years to look forward for comparison
                  backward_view=5  # minimum years to look back for comparison
                  ):
    cits = tt.cits()

    births = [tt.first]
    deaths = []

    last_life_avg = None

    for y_spl in range(tt.first + backward_view, tt.data_end + 1 - forward_view):
        lastb = births[-1]

        if len(births) > len(deaths):  # alive now
            if y_spl < births[-1] + backward_view:
                continue
            before = tt.sum_between(lastb, y_spl - 1) / (y_spl - lastb)
            after = tt.sum_between(y_spl, y_spl + forward_view - 1) / forward_view

            ratio = after / before
            # print(ratio)

            if ratio <= death_cutoff:  # died
                last_life_avg = before
                deaths.append(y_spl)

        else:  # dead now
            if y_spl < deaths[-1] + forward_view:  # need some time for a rebirth
                continue
            if tt.sum_between(y_spl, y_spl) == 0:  # don't be reborn in a year you get no citations
                continue

            after = tt.sum_between(y_spl, y_spl + forward_view - 1) / forward_view

            if after / last_life_avg > birth_cutoff:  # reborn
                births.append(y_spl)

    return (births, deaths)


from . import *

__all__ = [
    'yearly_counts_grid','matrix','key2name','yearly_counts_table',
    'author_pub_table'
]

def key2name(tname, truncate=None):
    tstr = tname.split("|")
    if len(tstr) == 3:
        tstr = "%s (%s)" % (tstr[0], tstr[1])
    else:
        tstr = "%s %s" % (tstr[0], tstr[1])
    tstr = tstr.title()

    if truncate is not None:
        if tstr[:truncate] != tstr:
            tstr = tstr[:truncate] + "..."

    return tstr

def _normalize(mat, axis=0):
    if axis==0: # rows
        return mat/mat.sum(axis=1)[:,None]
    elif axis==1: #columns
         return (mat.T/mat.T.sum(axis=1)[:,None]).T
    else:
        raise Exception('axis != 0/1')

# colormaps...
# https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html

def matrix(db, typ='docs', plot=False, norm=None, trans=False, includeComplement=False, **kwargs):

    assert(isinstance(db, Dataset))

    cnt = db.by( *list(kwargs) )

    # by1=range(35,50), by2=range(10,20)
    if len(list(kwargs)) != 2:
        raise Exception("matrix with dim != 2 not supported")

    if typ == 'docs':
        c = cnt.docs
    elif typ == 'cits':
        c = cnt.cits
    else:
        raise Exception("wtf... docs or cits bro")

    key1, key2 = sorted(kwargs)
    by1 = kwargs[key1]
    by2 = kwargs[key2]

    import numpy as np
    from collections import defaultdict

    def mk_array():
        for a1 in by1:
            my_series = np.array([c[(a1, a2)] for a2 in by2])

            if includeComplement:
                out = 0
                for (pp1, pp2), count in c.items():
                    # to account for a temporary bug
                    if pp1 < 0 or pp2 < 0:
                        continue

                    # skip what's already included
                    if pp1 != a1:
                        continue
                    if pp2 in by2:
                        continue

                    out += count

                my_series = np.array(list(my_series)+[out])

            yield my_series

        if includeComplement:
            cLast = defaultdict(int)

            outside_both = 0
            for (pp1, pp2),count in c.items():
                # to account for a temporary bug
                if pp1 < 0 or pp2 < 0:
                    continue

                # skip what's already included
                if pp1 in by1:
                    continue

                if pp2 not in by2:
                    outside_both += count

                #print(pp1, pp2, count)
                cLast[pp2] += count

            cLast = [cLast[x] for x in by2] + [outside_both]
            #print(cLast)
            yield cLast


    mat = np.array(list(mk_array()))

    if norm is not None:
        from . import matrix_normalize
        if norm not in kwargs:
            raise Exception('norm should be the axis along which you want to normalize')

        norm_i = sorted(kwargs).index(norm)
        mat = matrix_normalize(mat, axis=norm_i)

    if trans:
        mat = mat.transpose()

    if plot:
        from . import plt
        fig = plt.figure(figsize=(20, 4))  # plt.subplots()#
        plt.imshow(mat, interpolation='nearest', cmap=plt.cm.hsv)
        # IMS.cmap.set_under('yellow')
        # IMS.cmap.set_over('orange')
        plt.colorbar()

        if trans:
            xtick = by1
            ytick = by2
        else:
            xtick = by2
            ytick = by1

        tick_target = 5

        xstep = max(len(xtick) // tick_target,1)
        ystep = max(len(ytick) // tick_target,1)

        plt.xticks(
            range(0, len(xtick),xstep),
            xtick[::xstep],
        )
        plt.yticks(
            range(0, len(ytick),ystep),
            ytick[::ystep],
        )

        if trans:
            plt.xlabel(key1)
            plt.ylabel(key2)
        else:
            plt.xlabel(key2)
            plt.ylabel(key1)
    else:
        return mat

def yearly_counts_grid(names, dataset, myname=None,
    overwrite=True, 
    markers={}, print_names=None, 
    ctype='c', yearly_prop=False, 
    xlim=None, cols = 5,
    rows_per_group = 2,
        count_unit='doc', label_num='max'):

    assert(type(dataset) == Dataset)

    if xlim is None:
        xlim = (dataset.start, dataset.end)

    eps = (xlim[1]-xlim[0])/10
    xlim_display = (xlim[0]-eps, xlim[1]+eps)

    doing_both = False
    if type(count_unit) != str:
        if set(count_unit) == {'doc','cit'}:
            doing_both = True
        else:
            raise Exception('whoopsie')

    rows = len(names) // cols + bool(len(names) % cols) # = 15 for 5
    
    groupsize = rows_per_group * cols
    gs = rows // rows_per_group + bool(rows % rows_per_group)

    existing_files = Path(env.figure_dir).glob("%s.*" % myname)
    for fn in existing_files:
        fn.unlink()

    for groupi in range(gs):
        myFnName = "%s.%s" % (myname,groupi)
        if Path(myFnName).exists() and not overwrite:
            continue
        
        plt.figure(figsize=(cols*4,rows*1.2))
        plt.subplots_adjust(hspace=0.6, wspace=0)

        for i,name in enumerate(names[ groupi*groupsize : (groupi+1)*groupsize]):
            print(name)
            
            plt.subplot(rows,cols,i+1)

            yrstart,yrend = xlim
            years = range(yrstart, yrend+1)
            
            if count_unit == 'cit' or 'cit' in count_unit:
                vals = dataset.trend(ctype, name).cits(yrstart, yrend)
                
                if yearly_prop:
                    all_by_year = [ dataset(fy=YY).cits for YY in years ]
                    vals = np.array(vals) / np.array(all_by_year)

            else:
                vals = dataset.trend(ctype, name).docs(yrstart,yrend)
                
                if yearly_prop:
                    all_by_year = [ dataset(fy=YY).docs for YY in years ]
                    vals = np.array(vals) / np.array(all_by_year)

            # fill that shit up
            plt.fill_between(years,vals,color='black',alpha=0.4)


            if doing_both:
                vals2 = np.array(dataset.trend(ctype, name).docs())
                
                if yearly_prop:
                    all_by_year = [ dataset(fy=YY).docs for YY in years ]
                    vals2 = np.array(vals2) / np.array(all_by_year)
                    
                # this one is scaled...
                vals2 = np.max(vals) * vals2 * 0.5 / np.max(vals2)

                plt.plot(years, vals2, color='black')
            
            if print_names is not None:
                title = print_names(name)
            else:
                title = name
                
            if len(title) >= 40:
                title = title[:37]+"..."
                
            #t = plt.text(min(years) + (max(years)-min(years))/2, 1.2*(max(vals))/1,title, fontsize=13)
            plt.title(title)
            
            plt.axis('off')

            from math import ceil, floor

            XAXIS_YSHIFT = -max(vals)/25
            lines = []

            if False:
                # lines on the x-axis...
                for decade in range(floor(yrstart/10)*10,ceil(yrend/10)*10,10):
                    lines += [
                        (decade+1, decade+10-1), 
                        #(-max(vals)/5, -max(vals)/5), 
                        #(0,0),
                        (XAXIS_YSHIFT, XAXIS_YSHIFT), 
                        'black'
                    ]

            # labeling the last
            if label_num == 'last':
                last_num = vals[-1]

                if yearly_prop:
                    last_nump = "%0.1f%%"%(last_num*100)
                else:
                    last_nump = last_num

                lines += [
                    (xlim_display[1]-10, xlim_display[1]-5),
                    (last_num, last_num),
                    "black"
                ]
                plt.text(xlim_display[1]-3, last_num, last_nump, fontsize=12, verticalalignment='center', horizontalalignment='left')#*3+min(vals)
                
            elif label_num == 'max':
                #print(vals, title)
                max_num = np.max(vals)

                if yearly_prop:
                    max_nump = "%0.1f%%"%(max_num*100)
                else:
                    max_nump = max_num

                lines += [
                    (xlim_display[1]-10, xlim_display[1]-5),
                    (max_num, max_num),
                    "black"
                ]
                plt.text(xlim_display[1]-3, max_num, max_nump, fontsize=12, verticalalignment='center', horizontalalignment='left')#*3+min(vals)

            elif label_num is not None:
                raise Exception('don\'t recognize that label_num...')

            plt.plot(*lines)

            # dots on the x-axis
            half_centuries = range( ceil(xlim[0]/50)*50, floor(xlim[1]/50)*50 + 50, 50 )
            decades = list(set(range( ceil(xlim[0]/10)*10, floor(xlim[1]/10)*10 + 10, 10 )).difference(set(half_centuries)))
            plt.scatter(
                decades,
                [0]*len(decades),
                color='black',
                s=5
            )

            #print('max',np.max(vals))
            #print(list(vals))
            if True:
                plt.vlines(
                    half_centuries,
                    -np.max(vals)/5,
                    0
                )
            
            plt.xlim(xlim_display)
            plt.ylim((-np.max(vals)/5,np.max(vals)))
            
            # put the markers on the time seriesss
            if name in markers:
                ms = markers[ name ]
                if type(ms) == list:
                    #print('here!')
                    plt.scatter(
                        ms,
                        [-max(vals)/10]*len(ms),
                        color='red',
                        s=30,
                        marker="^"
                    )
                elif type(ms) == dict:
                    for k,v in ms.items():
                        plt.text(k, -max(vals)/2, v)

        dataset.save_figure(myFnName)
        plt.show();



def yearly_counts_table_simp(
        dta, who,
        NCOLS=2, NPAGES=None,
        markranges={}, yearlim=(1950,2015),
        tickstep=20,
        print_names={}, number_start=1,
        print_numbers = False):

    wper = 16 / 3
    hper = 14 / 20
    NROWS = (len(who) // NCOLS) + int(len(who) % NCOLS != 0)

    # top_end = (2080-2015) /  (2015-1950)
    # bottom_end = (1950 - 1880) / (2015 - 1950)
    # year_wid = yearlim[1] - yearlim[0]
    yearmin, yearmax = yearlim

    # xmin, xmax = (yearlim[0] - bottom_end * year_wid, yearlim[1] + top_end * year_wid)
    # def dist_trans(nyears):
    #    return year_wid * nyears / (2015-1950)

    plt.figure(figsize=(wper * NCOLS, hper * NROWS))

    for col_i in range(NCOLS):

        plt.subplot(1, NCOLS, col_i + 1)

        trends = []
        for i, tname in enumerate(who[col_i * NROWS: (col_i + 1) * NROWS]):
            tt = np.array(dta.trend('c', tname).cits(yearmin, yearmax))

            # print(yearmin,yearmax)
            trends.append((tname, tt))

        for i, (tname, tt) in enumerate(trends):
            ttm = tt.max()
            ttl = tt[-1]
            ttl15 = tt[-15:].mean()

            BASE = (NROWS - 1 - i) * 2

            tt = tt / ttm
            tt = tt + BASE

            years = range(yearmin, yearmax + 1)

            plt.axis('off')
            plt.fill_between(years, tt, y2=BASE, color='gray')

            if tname in print_names:
                tstr = print_names[tname]
            else:
                tstr = key2name(tname, truncate=25)

            tstr = "(%s) " % ((col_i) * NROWS + i + number_start) * print_numbers + tstr
            plt.text(2020, BASE, tstr, fontsize=13)
            # plt.text(yearmax + dist_trans(2030-2015), BASE, int(ttm), horizontalalignment='center')
            # plt.text(yearmax + dist_trans(2050-2015), BASE, int(ttl), horizontalalignment='center')
            # plt.text(yearmax + dist_trans(2070-2015), BASE, "%0.1f" % ttl15, horizontalalignment='center')
            # plt.text(2025, i*2, int(ttm), horizontalalignment='center')

            if tname in markranges:
                # xmin,xmax = plt.xlim()
                ymin, ymax = [BASE, BASE + 1.4]
                mk = markranges[tname]
                plt.fill_between(mk, [ymin, ymin], [ymax, ymax], color='blue', alpha=0.2)

        plt.vlines(range(yearmin, yearmax, tickstep), 0, NROWS * 2, color='black', alpha=0.3)

        # table lines...
        # plt.vlines([ yearmax + dist_trans(DD-2015) for DD in range(2020, 2080 + 20, 20)], 0, NROWS * 2, color='black', alpha=1)  # table lines

        for y in range(yearmin, yearmax, tickstep):
            plt.text(y, -1, y, horizontalalignment='center')

        plt.xlim(yearmin, yearmax + 50)


def yearly_counts_table(
        dta, who,
        NCOLS=2, NPAGES=None,
        markranges={}, yearlim=(1950,2015),
        tickstep=20,
        print_names={}, number_start=1
):
    if NPAGES is not None:
        nper = len(who) // NPAGES + int( len(who) % NPAGES != 0 )
        for i in range(NPAGES):
            yearly_counts_table(
                dta,
                who[i*nper:(i+1)*nper],
                NCOLS,None,
                markranges,
                yearlim,
                tickstep,
                print_names=print_names, number_start=number_start
            )
        return

    #print(yearlim)

    wper = 17 / 3
    hper = 10 / 20
    NROWS = (len(who) // NCOLS) + int(len(who) % NCOLS != 0)

    top_end = (2080-2015) /  (2015-1950)
    bottom_end = (1950 - 1850) / (2015 - 1950)
    year_wid = yearlim[1] - yearlim[0]
    yearmin, yearmax = yearlim

    xmin, xmax = (yearlim[0] - bottom_end * year_wid, yearlim[1] + top_end * year_wid)
    def dist_trans(nyears):
        return year_wid * nyears / (2015-1950)


    plt.figure(figsize=(wper * NCOLS, hper * NROWS))

    for col_i in range(NCOLS):

        plt.subplot(1, NCOLS, col_i + 1)

        trends = []
        for i, tname in enumerate(who[col_i * NROWS: (col_i + 1) * NROWS]):
            tt = np.array(dta.trend('c', tname).cits(yearmin, yearmax))

            #print(yearmin,yearmax)
            trends.append((tname, tt))

        for i, (tname, tt) in enumerate(trends):
            ttm = tt.max()
            ttl = tt[-1]
            ttl15 = tt[-15:].mean()

            BASE = (NROWS - 1 - i) * 2

            tt = tt / ttm
            tt = tt + BASE

            years = range(yearmin, yearmax+1)

            plt.axis('off')
            plt.fill_between(years, tt, y2=BASE, color='gray')

            if tname in print_names:
                tstr = print_names[tname]
            else:
                tstr = key2name(tname, truncate=25)

            tstr = "(%s) %s" % ((col_i) * NROWS + i + number_start, tstr)
            plt.text(yearmin - dist_trans(1950-1820), BASE, tstr)
            plt.text(yearmax + dist_trans(2030-2015), BASE, int(ttm), horizontalalignment='center')
            plt.text(yearmax + dist_trans(2050-2015), BASE, int(ttl), horizontalalignment='center')
            plt.text(yearmax + dist_trans(2070-2015), BASE, "%0.1f" % ttl15, horizontalalignment='center')
            # plt.text(2025, i*2, int(ttm), horizontalalignment='center')

            if tname in markranges:
                #xmin,xmax = plt.xlim()
                ymin,ymax = [BASE,BASE+1.4]
                mk = markranges[tname]
                plt.fill_between( mk, [ymin,ymin], [ymax,ymax], color='blue', alpha=0.2 )

        plt.vlines(range(yearmin, yearmax, tickstep), 0, NROWS * 2, color='black', alpha=0.3)

        # table lines...
        plt.vlines([ yearmax + dist_trans(DD-2015) for DD in range(2020, 2080 + 20, 20)], 0, NROWS * 2, color='black', alpha=1)  # table lines

        for y in range(yearmin, yearmax, tickstep):
            plt.text(y, -1, y, horizontalalignment='center')

        plt.xlim(xmin, xmax)

    # plt.savefig("top_cited.png", dpi=300)

def author_pub_table(dta, who, **kwargs):
    yearly_counts_table(dta, dta.search('c', who + "|"), **kwargs)




__all__ = [
    'figure_dir',
    'GLOBS','cfile'
]

import os
from pathlib import Path
import yaml

#print(os.environ)

# more local stuff
if  'NBDIR' in os.environ:
    notebook_dir = Path(os.environ['NBDIR']).parent
    figure_dir = notebook_dir.joinpath('figures')
    #table_dir = notebook_dir.joinpath('tables')


    for d in [figure_dir]:#,table_dir]:
        if not d.exists():
            print("Creating directory: %s" % d)
            d.mkdir()

cfile = Path(__file__).parent.joinpath('config.yaml')

if cfile.exists():
    with cfile.open('r') as f:
        GLOBS = yaml.load(f, Loader=yaml.FullLoader)
else:
    GLOBS = {}

if 'kkdir' in GLOBS:
    #variable_dir = Path(os.environ['VARDIR']) if 'VARDIR' in os.environ else None
    variable_dir = Path(GLOBS['kkdir'], 'data')


from peewee import *
from .. import env

db = SqliteDatabase( env.variable_dir.joinpath('wos.db') )
#db = PostgresqlDatabase('wos', user='postgres', password='postgres',
#                           host='0.0.0.0', port=5432)

class JSONField(TextField):
    """Store JSON data in a TextField."""
    def python_value(self, value):
        if value is not None:
            return json.loads(value)

    def db_value(self, value):
        if value is not None:
            return json.dumps(value)

class Doc(Model):
    title = TextField()
    first_author = CharField(index=True)
    journal = CharField(index=True)
    year = IntegerField(index=True)
    citcnt_external = IntegerField()

    class Meta:
        database = db

class Cit(Model):
    doc = ForeignKeyField(Doc, backref='citations')
    full_ref = CharField(index=True)
    author = CharField(index=True)
    year = IntegerField(index=True)
    
    class Meta:
        database = db

class Keyword(Model):
    doc = ForeignKeyField(Doc, backref='keywords')
    text = TextField(index=True)

    class Meta:
        database = db


from .. import *
from ..nlp import *

#from jstor_xml import *
#from intext import *

from nltk import sent_tokenize
from lxml.etree import _ElementTree as ElementTree
from lxml import etree
recovering_parser = etree.XMLParser(recover=True)

from zipfile import ZipFile





if False:
    # looks like I don't use this?
    import spacy

    nlp = spacy.load('en')
    import neuralcoref

    neuralcoref.add_to_pipe(nlp)





ALL_UNIVERSITIES = {
    x[1].upper().strip()
    for x in creader(
        Path(__file__).parent.parent.joinpath("external-data", "world-universities.csv")
            .open('r', encoding='UTF8')
    )
}
max_uni_length = max( len(x) for x in ALL_UNIVERSITIES )







def citation_iterator(t):
    citations = extractCitation(t)
    for c in citations:
        name = c['names']
        name = name.replace("'s", "")  # change "Simmel's" to "Simmel"

        years = c['year'].split(",")
        for y in years:
            y = y.strip()  # get rid of the extra spacing
            y = re.findall(r'[0-9]{4}', y)  # this strips out the 'a' in '1995a'
            if not len(y):
                faulty_years.add(c['year'])
                continue

            y = y[0]
            yield "%s (%s)" % (name, y)


class ParseError(Exception):
    pass

def clean_metadata( doi, metadata_str ):

    try:
        metaXml = etree.fromstring(metadata_str, parser=recovering_parser)
    except OSError:
        raise ParseError("NO METADATA!", doi)

    def findAndCombine(query):
        return ";".join([" ".join(x.itertext()) for x in metaXml.findall(query)]).strip()

    metadata = {}
    metadata['type'] = metaXml.get("article-type")

    metadata['doi'] = doi
    # print(doi)

    metadata['title'] = findAndCombine(".//article-title")
    metadata['journal'] = findAndCombine(".//journal-title")
    metadata['publisher'] = findAndCombine(".//publisher-name")

    metadata['abstract'] = findAndCombine(".//article-meta//abstract")

    auth = []
    for group in metaXml.findall(".//contrib"):
        myname = " ".join(x.strip() for x in group.itertext())
        myname = " ".join(re.split("\s+", myname)).strip()
        auth.append(myname)
    metadata['authors'] = auth

    if len(auth) == 0 and False:
        print(doi)
        print(auth)
        print(metadata['title'])

    metadata['month'] = findAndCombine(".//article-meta//month")

    metadata['year'] = findAndCombine(".//article-meta//year")
    if ";" in metadata['year']:
        metadata['year'] = metadata['year'].split(";")[0]
    try:
        metadata['year'] = int(metadata['year'])
    except:
        raise ParseError("No valid year found")

    metadata['volume'] = findAndCombine(".//article-meta//volume")
    metadata['issue'] = findAndCombine(".//article-meta//issue")

    metadata['fpage'] = findAndCombine(".//article-meta//fpage")
    metadata['lpage'] = findAndCombine(".//article-meta//lpage")

    return metadata


def basic_ocr_cleaning(x):
    # remove multiple spaces in a row
    x = re.sub(r" +", ' ', str(x))
    # remove hyphenations [NOTE this should be updated, with respect to header and footer across pages...]
    x = re.sub(r"([A-Za-z]+)-\s+([A-Za-z]+)", "\g<1>\g<2>", x)

    x = x.strip()
    return x


def get_content_string(ocr_string):
    docXml = etree.fromstring(ocr_string, parser=recovering_parser)
    pages = docXml.findall(".//page")

    page_strings = []
    for p in pages:
        if p.text is None:
            continue
        page_strings.append(p.text)

    secs = docXml.findall(".//sec")

    for s in secs:
        if s.text is None:
            continue
        if s.text.strip() == '':
            try_another = etree.tostring(s, encoding='utf8', method='text').decode("utf8").strip()
            #print(try_another)
            if try_another == '':
                continue

            page_strings.append(try_another)
        else:
            page_strings.append(s.text.strip())

    return basic_ocr_cleaning( "\n\n".join(page_strings) )








class Sentence:
    def __init__(self, content=None ,lines=None, words=None):
        self.lines = lines
        self.content = content
        self.words = words
        self.type = None

        if self.words is not None:
            self.content = untokenize(self.words)

        # process hyphenation
        self.content = fix_paragraph(self.content)

    def parse(self):

        self.doc = nlp(self.content)
        self.root_verb = [word for word in self.doc if word.head == word ][0]

        ".+ is one thing and .+ is quite another"
        ".+ is one thing and .+ is another"
        ".+ is the same as .+"

        if str(self.root_verb) == "is":

            nsubj = follow_expand( self.root_verb, 'nsubj' )
            attr = follow_expand( self.root_verb, 'attr' )
            advc = follow_expand( self.root_verb, 'advcl' )

            if len(nsubj) == 1 and len(attr) == 1:
                self.type = "identification"
                self.args = [
                    nsubj[0],
                    attr[0],
                    advc
                ]
                print("X(%s) IS Y(%s)" % tuple(self.args[:2]))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.content

class Line:
    def __init__(self, content=None, page=None):
        self.content = content.strip()
        self.page = None
        self.type = None

        #if len(self.content) < 5:
        #    print(self.content)

        if self.content == "":
            self.type = "empty"

        try:
            numerical_content = int(self.content)
            self.type = "page_number"
            self.value = numerical_content
        except ValueError:
            pass

    def __str__(self):
        return self.__repr__()
    def __repr__(self):
        return self.content

class Page:

    @classmethod
    def from_lines(cls, lines):
        p = Page()

        for line in lines:
            p.lines.append(Line(
                # content
                content=line,
                # and context
                page=p
            ))

        p.extract_sentences()

        return p

    @classmethod
    def from_text(cls, text):
        return Page.from_lines(re.split("[\n\r]+", text))

    def __init__(self):
        self.lines = []
        self.start_stub = None
        self.end_stub = None
        self.full_sentences = None


    def __str__(self):
        return self.__repr__()
    def __repr__(self):
        return "\n".join( map(str,self.lines) )

    def extract_sentences(self):
        from nltk import sent_tokenize, word_tokenize

        sents = [ word_tokenize(ss) for ss in sent_tokenize( str(self) ) ]

        if not len(sents):
            raise ParseError("Sentences not found")

        full_sentences = [ Sentence(words=sent_words) for sent_words in sents[1:-1] ]
        start_stub = sents[0]
        end_stub = sents[-1]

        self.start_stub = start_stub
        self.end_stub = end_stub
        self.full_sentences = full_sentences

    @property
    def content(self):
        return str(self)

romanChars = "LIXMCV"

def isRoman(x):
    if not type(x) == str:
        return False

    return all( y in romanChars for y in x.upper() )

def num_pdf2page(pdf_num, page_map):
    import roman
    beforeMe = [ x for x in page_map.items() if x[0] <= pdf_num ]
    if not len(beforeMe):
        return -1
    lastBeforeMe = max( beforeMe, key=lambda x: x[0] )

    afterMe = [x for x in page_map.items() if x[0] >= pdf_num]
    if not len(afterMe):
        return -1
    firstAfterMe = min(afterMe, key=lambda x: x[0])

    if isRoman(lastBeforeMe[1]) and isRoman(firstAfterMe[1]):
        return roman.toRoman( pdf_num - lastBeforeMe[0] + roman.fromRoman(lastBeforeMe[1].upper()) )

    elif not isRoman(lastBeforeMe[1]) and not isRoman(firstAfterMe[1]):
        return pdf_num - lastBeforeMe[0] + int(lastBeforeMe[1])

    else:
        return -1





class Document:
    def __init__(self, metadata):
        self.pages = []
        self.metadata = metadata

        self.type = self.metadata['type']

        lt = self['title'].lower()
        if lt in ["book review", 'review essay']:
            self.type = 'book-review'
        if "Commentary and Debate".lower() == lt or 'Comment on'.lower() in lt:
            self.type = 'commentary'
        if lt.find('reply to') == 0:
            self.type = 'commentary'
        if 'Erratum'.lower() in lt:
            self.type = 'erratum'

    @classmethod
    def from_db(cls, _id):
        raise Exception("Haven't done yet!")

    @classmethod
    def from_xml_strings(cls, meta_xml, content_xml, doi=None, complex_parsing=True):

        metadata = cls.parse_metadata(meta_xml)
        metadata['doi'] = doi
        page_strings = cls.parse_content(content_xml)

        page_strings = list(map(basic_ocr_cleaning, page_strings))
        page_strings = [x for x in page_strings if x != ""]
        if not len(page_strings):
            raise ParseError("Empty document")

        return Document.from_pages(page_strings, metadata=metadata, complex_parsing=complex_parsing)


    @classmethod
    def from_file(cls, fn, complex_parsing=True):

        my_name = ".".join(fn.split(".")[:-1])
        doi = my_name.split("-")[-1].replace("_", "/")

        metadataFn = "%s.xml" % my_name
        metadataFn = join(dirname(metadataFn), "..", "metadata", metadataFn)

        metadata = cls.parse_metadata(open(metadataFn).read())


        page_strings = cls.parse_content(open(fn).read())

        page_strings = list(map(basic_ocr_cleaning, page_strings))
        page_strings = [x for x in page_strings if x != ""]
        if not len(page_strings):
            raise ParseError("Empty document")

        return Document.from_pages(page_strings, metadata=metadata, complex_parsing=complex_parsing)

    @classmethod
    def from_pages(cls, page_strings, metadata={}, complex_parsing=True):
        d = Document(metadata)

        if not len(page_strings):
            raise ParseError("No pages...")

        d.pages = [Page.from_lines(re.split(r"[\n\r]+", page)) for page in page_strings]

        if complex_parsing:
            if d.type == 'research-article':
                try:
                    d.find_bibliography()
                    d.extract_headers_footers()
                    #d.parse_bibliography()
                except ParseError:
                    print("Couldn't extract bib and headers from ", d['doi'])
                    raise

        return d

    @property
    def sentences(self):
        for i, p in enumerate(self.pages):
            if i > 0:
                yield Sentence(words=self.pages[i - 1].end_stub + self.pages[i].start_stub)

            p.extract_sentences()
            for s in p.full_sentences:
                yield s

    @classmethod
    def parse_content(cls, xml_string):

        import xml.etree.ElementTree as ET
        docXml = ET.ElementTree(ET.fromstring(xml_string, parser=recovering_parser))

        pages = docXml.findall(".//page")

        page_strings = []
        for p in pages:
            if p.text is None:
                continue
            page_strings.append(p.text)

        secs = docXml.findall(".//sec")

        for s in secs:
            if s.text is None:
                continue
            if s.text.strip() == '':
                try_another = etree.tostring(s, encoding='utf8', method='text').decode("utf8").strip()
                # print(try_another)
                if try_another == '':
                    continue

                page_strings.append(try_another)
            else:
                page_strings.append(s.text.strip())

        return page_strings

    @classmethod
    def parse_metadata(cls, xml_string):

        try:
            import xml.etree.ElementTree as ET
            metaXml = ET.ElementTree(ET.fromstring(xml_string, parser=recovering_parser))
        except OSError:
            raise ParseError("NO METADATA!", fn)

        assert (isinstance(metaXml, ET.ElementTree))

        def findAndCombine(query):
            return ";".join([" ".join(x.itertext()) for x in metaXml.findall(query)]).strip()

        metadata = {}
        metadata['type'] = metaXml.getroot().get("article-type")

        # print(doi)

        metadata['title'] = findAndCombine(".//article-title")
        metadata['journal'] = findAndCombine(".//journal-title")
        metadata['publisher'] = findAndCombine(".//publisher-name")

        metadata['abstract'] = findAndCombine(".//article-meta//abstract")

        auth = []
        for group in metaXml.findall(".//contrib"):
            myname = " ".join(x.strip() for x in group.itertext())
            myname = " ".join(re.split("\s+", myname)).strip()
            auth.append(myname)
        metadata['authors'] = auth

        if len(auth) == 0 and False:
            print('no authors...')
            print(auth)
            print(metadata['title'])

        metadata['month'] = findAndCombine(".//article-meta//month")

        metadata['year'] = findAndCombine(".//article-meta//year")
        if ";" in metadata['year']:
            metadata['year'] = metadata['year'].split(";")[0]
        try:
            metadata['year'] = int(metadata['year'])
        except:
            raise ParseError("No valid year found")

        metadata['volume'] = findAndCombine(".//article-meta//volume")
        metadata['issue'] = findAndCombine(".//article-meta//issue")

        metadata['fpage'] = findAndCombine(".//article-meta//fpage")
        metadata['lpage'] = findAndCombine(".//article-meta//lpage")

        return metadata

    @property
    def num_pages(self):
        return len(self.pages)

    def apply_page_map(self, page_map):
        self.page_map = page_map

        for i, p in enumerate(self.pages):
            p.page_mapped = num_pdf2page(pdf_num=i, page_map=self.page_map)

    def eliminate_metadata(self):
        from fuzzysearch import find_near_matches

        if self.type != 'research-article':
            print("Unimplemented: extracting headers and footers on non-research-article")
            return

        if not len(self.pages):
            raise ParseError()

        fp = str(self.pages[0])
        fp = fp.upper()

        # extracting title, heading stuff on first page, and abstract
        to_extract_from_first_page = [
            self['title']
        ]

        if len(self['abstract']):
            to_extract_from_first_page.append(self['abstract'])

        for a in self['authors']:
            arev = a.split()
            arev = arev[-1] + ", " + " ".join(arev[:-1])

            to_extract_from_first_page += [a, arev]

        found = []

        # to get the university, not included in most metadatas...
        if False:
            found_uni = []
            for u in ALL_UNIVERSITIES:
                found.append((fp.find(u), len(u)))
        print(to_extract_from_first_page)

        for word in to_extract_from_first_page:
            for match in find_near_matches(word.upper(), fp, max_l_dist=len(word) // 10):
                # match = word[match.start:match.end]
                # print('match {}'.format(match))
                # index = ls.find(match)
                print(match.start, match.end, match.matched, match.dist)
                found.append( (match.start, match.end-match.start) )



        # limit to those strings which were actually found
        found = list(filter(lambda x: x[0] >= 0, found))

        if len(found):
            # these strings should be "end to end", i.e. less than 5 characters separating them
            found_ends = [x[0] + x[1] for x in found]
            found_min_beginning = min(x[0] for x in found)
            found = list(
                filter(lambda x: any(abs(x[0] - y) < 5 for y in found_ends) or x[0] == found_min_beginning, found)
            )

            # we then cut after these metadata fields!
            doc_start = max(found, key=lambda x: x[0] + x[1])
            doc_start = doc_start[0] + doc_start[1]

            newfp_text = str(self.pages[0])[doc_start:].strip()
            # newfp = Page.from_text(newfp_text)
            self.pages[0] = newfp_text
            #print("Trimmed woohoo: %s\n\n%s"%(fp, newfp_text))

    def extract_headers_footers(self):
        # extracting headers via voting
        num_identical_necessary = len(self.pages) / 4

        flines = [str(x).split()[:20] for x in self.pages]
        grams = Counter(
            tuple(x[:i])
            for i in range(1, 15)
            for x in flines
        )

        candidates = set(x for x, c in grams.items() if c > num_identical_necessary)
        maximal_candidates = []
        for c in candidates:
            keep = True
            for c2 in candidates:
                if len(c) >= len(c2):
                    continue
                if c[:min(len(c), len(c2))] == c2[:min(len(c), len(c2))]:
                    keep = False

            if keep:
                maximal_candidates.append(c)

        self.headers = [" ".join(x) for x in maximal_candidates]
        for h in self.headers:

            # don't want to get rid of the title on the first page.
            # that's the job for another algorithm
            # therefore, the [1:]
            for i, p in enumerate(self.pages[1:]):
                try:
                    if str(p).index(h) < 30:
                        self.pages[i+1] = Page.from_text(str(p).replace(h, "").strip())
                except ValueError:
                    continue

    def parse_bibliography(self):
        if self.type != 'research-article':
            print("Unimplemented: parsing bibliography on non-research-article")
            self.bibliography = []
            return

        from citation_parser import citation_grammars
        from citation_parser import CitationVisitor
        from parsimonious.exceptions import ParseError
        cv = CitationVisitor()

        def name_split(x):

            # go backwards, finding the last colon or number. then forwards to the next period.
            parts = re.split(r"([:][^:.]*\.|[0-9][^0-9.]*\.)", x)
            parts = ["".join(parts[:-1]), parts[-1]]

            return parts

        t = self.bibString
        # split by years first, to find separate citations
        yspl = re.split("((?:19|20)[0-9]{2}[^\-])", t)
        print("\n".join("--%s" % x for x in yspl))
        bib_entries = []
        for i in range(1, len(yspl), 2):
            # lands on every year.
            if i == 1:
                bib_entries.append("".join([
                    yspl[0],  # stub at beginning
                    yspl[1],  # year that was split
                    name_split(yspl[2])[0]  # split off the name from the bulk of the citation
                ]).strip())
                continue

            bib_entries.append("".join([
                name_split(yspl[i - 1])[1],  # name of the last one
                yspl[i],  # year that was split
                name_split(yspl[i + 1])[0]  # split off the name from the bulk of the citation
            ]).strip())

        print("\n".join("--%s" % x for x in bib_entries))

        errors = 0
        success = 0
        last_success = None

        parsed_bibs = []

        for bib_entry in bib_entries:
            # print("Parsing %s" % bib_entry)

            # this block of code tries to replace ------ at the beginning of the line
            # with the most recent successfull parse
            if last_success is not None:
                a = last_success['authors']
                fa = list(a[0])
                fa = fa[-1] + ", " + " ".join(fa[:-1])

                a = [" ".join(x) for x in a]

                if len(a) > 1:
                    author_string = ", ".join(
                        [fa] +
                        a[1:-1] +
                        ["and " + a[-1]]
                    )
                else:
                    author_string = fa

                bib_entry = re.sub("^[-–—]+", author_string, bib_entry)

            # now we throw it into the parser.
            # not everything works, but it does a pretty good job I think, for now.
            for cg in citation_grammars:
                try:
                    p = cg.parse(bib_entry)
                    p = cv.visit(p)
                    p = p[1][0]

                    last_success = p
                    parsed_bibs.append(p)
                    success += 1
                    break
                except ParseError:
                    errors += 1
                    last_success = None

        self.bibliography = parsed_bibs

    def find_bibliography(self):
        if self.type != 'research-article':
            print("Unimplemented: finding bibliography on non-research-article")
            return

        num = 0
        pid = None
        for pi, p in enumerate(self.pages):
            fr = re.findall("(REFERENCES|References|Literature Cited)", str(p))
            if not len(fr):
                continue
            num += len(fr)
            pid = pi

        if num > 1:
            #raise ParseError("More than one REFERENCES string!")
            self.bibString = ""
            return

        if pid is None:
            #raise ParseError("No REFERENCE string found...", "\n".join( str(p) for p in self.pages ))
            self.bibString = ""
            return

        # print("Found bibliography starting on page ", pid+1)

        bibString = []
        for i in range(pid, len(self.pages)):
            if i == pid:
                bibString.append("".join(re.split("(REFERENCES|References|Literature Cited)", str(self.pages[i]))[2:]))
                continue

            bibString.append(str(self.pages[i]))

        self.bibString = "\n".join(bibString).strip()

        newptext = str(self.pages[pid]).split("REFERENCES")[0].strip()
        if newptext == "":
            # print("Deleting page", pid+1)
            self.pages.pop()
        else:
            # print("Truncating page", pid+1)
            self.pages[pid] = Page.from_text(newptext)

        # print("Deleting pages ", pid+2, "through", len(self.pages))
        for i in range(len(self.pages) - pid):
            self.pages.pop()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "\n\n".join(map(str, self.pages))

    def __getitem__(self, item):
        return self.metadata[item]


_zp_cache = {}


def doi2xml(zipdir, doi):
    zf, fn, fn2 = doi2fn(zipdir, doi)

    archive = ZipFile(zf, 'r')
    return (
        archive.read(fn),
        archive.read(fn2)
    )

def load_zd(zipdir):
        zipfiles = list(Path(zipdir).glob("*.zip"))

        zret = defaultdict(set)

        for zf in zipfiles:
            archive = ZipFile(zf, 'r')
            files = archive.namelist()

            for fn in files:
                if 'ocr/' in fn or 'ngram/' in fn:
                    continue

                fn2 = fn.replace("metadata", "ocr").replace(".xml", ".txt")

                mydoi = fn2doi(fn)
                zret[ mydoi ] = (zf, fn, fn2)

        _zp_cache[zipdir] = zret

def doi2fn(zipdir, doi):
    fn = "journal-article-%s.txt" % doi.replace("/", "_")

    if zipdir not in _zp_cache:
        load_zd(zipdir)

    return _zp_cache[zipdir][ doi ]


def fn2doi(fn):
    my_name = ".".join(fn.split(".")[:-1])
    doi = my_name.split("-")[-1].replace("_", "/")
    return doi












def _zip_doc_iterator(zipfiles):
    from random import shuffle

    all_files = []
    for zf in zipfiles:
        archive = ZipFile(zf, 'r')
        files = archive.namelist()
        names = list(set(getname(x) for x in files))

        all_files += [(archive, name) for name in names]

    shuffle(all_files)

    for archive, name in all_files:
        try:
            yield (
                name.split("-")[-1].replace("_", "/"),
                archive.read("metadata/%s.xml" % name),
                archive.read("ocr/%s.txt" % name).decode('utf8')
            )
        except KeyError:  # some very few articles don't have both
            continue




# some types of titles should be immediately ignored
def title_looks_researchy(lt):
    lt = lt.lower()
    lt = lt.strip()

    for x in ["book review", 'review essay', 'back matter', 'front matter', 'notes for contributors',
              'publication received', 'errata:', 'erratum:']:
        if x in lt:
            return False

    for x in ["commentary and debate", 'erratum', '']:
        if x == lt:
            return False

    return True



def doc_iterator(jstor_zip_base, research_limit=True, complex_parsing=True):
    zipfiles = list(Path(jstor_zip_base).glob("*.zip"))

    for i, (doi, metadata_str, ocr_str) in enumerate(_zip_doc_iterator(zipfiles)):

        try:
            drep = Document.parse_metadata(metadata_str)

            if research_limit:
                if drep['type'] != 'research-article':
                    continue

                lt = drep['title'].lower()
                if not title_looks_researchy(lt):
                    continue

                # Don't process the document if there are no authors
                if not len(drep['authors']):
                    continue

            d = Document.from_xml_strings(metadata_str, ocr_str, complex_parsing=complex_parsing)
            d.metadata['doi'] = doi
        except ParseError:
            continue

        yield d


def doc_iterator_forMongo(jstor_zip_base,
                 journals_filter=None,
                 debug=False):

    zipfiles = list(Path(jstor_zip_base).glob("*.zip"))

    for i, (doi, metadata_str, ocr_str) in enumerate(_zip_doc_iterator(zipfiles)):
        try:
            drep = clean_metadata(doi, metadata_str)

            # only include journals in the list "included_journals"
            if journals_filter is not None and (drep['journal'].lower() not in journals_filter):
                continue

            if debug: print("got meta")

            if drep['type'] != 'research-article':
                continue

            # some types of titles should be immediately ignored
            def title_looks_researchy(lt):
                lt = lt.lower()
                lt = lt.strip()

                for x in ["book review", 'review essay', 'back matter', 'front matter', 'notes for contributors',
                          'publication received', 'errata:', 'erratum:']:
                    if x in lt:
                        return False

                for x in ["commentary and debate", 'erratum', '']:
                    if x == lt:
                        return False

                return True

            lt = drep['title'].lower()
            if not title_looks_researchy(lt):
                continue

            # Don't process the document if there are no authors
            if not len(drep['authors']):
                continue

            drep['content'] = get_content_string(ocr_str)

            drep['citations'] = []

            # loop through the matching parentheses in the document
            for index, (parenStart, parenContents) in enumerate(getOuterParens(drep['content'])):

                citations = list(citation_iterator(parenContents))
                if not len(citations):
                    continue

                citation = {
                    "citations": citations,
                    "contextLeft": drep['content'][parenStart - 400 + 1:parenStart + 1],
                    "contextRight": drep['content'][
                                    parenStart + len(parenContents) + 1:parenStart + len(parenContents) + 1 + 100],
                    "where": parenStart
                }

                # cut off any stuff before the first space
                first_break_left = re.search(r"[\s.!?]+", citation['contextLeft'])
                if first_break_left is not None:
                    clean_start_left = citation['contextLeft'][first_break_left.end():]
                else:
                    clean_start_left = citation['contextLeft']

                # cut off any stuff after the last space
                last_break_right = list(re.finditer(r"[\s.!?]+", citation['contextRight']))
                if len(last_break_right):
                    clean_end_right = citation['contextRight'][:last_break_right[-1].start()]
                else:
                    clean_end_right = citation['contextRight']

                # we don't want anything more than a sentence

                sentence_left = sent_tokenize(clean_start_left)
                if len(sentence_left):
                    sentence_left = sentence_left[-1]
                else:
                    sentence_left = ""

                sentence_right = sent_tokenize(clean_end_right)[0]
                if len(sentence_right):
                    sentence_right = sentence_right[0]
                else:
                    sentence_right = ""

                # finally, strip the parentheses from the string
                sentence_left = sentence_left[:-1]
                sentence_right = sentence_right[1:]

                # add the thing in context
                full = sentence_left + "<CITATION>" + sentence_right

                citation['contextPure'] = sentence_left
                # print(full)

                drep['citations'].append(citation)

            yield doi, drep

        except ParseError as e:
            print("parse error...", e.args, doi)










class jstor_counter:

    def __init__(
            self,
            jstor_zip_base, output_database, name_blacklist=[],
            RUN_EVERYTHING=False,
            groups=None, group_reps=None,
            citations_filter=None, journals_filter=None, debug=False,
            CONSOLIDATE_ITERS=0, term_whitelist = set(), NUM_TERMS_TO_KEEP=100, stopwords = None
    ):

        if stopwords is None:
            # getting ready for term counting
            from nltk.corpus import stopwords as sw
            self.stopwords = set(sw.words('english'))
        else:
            self.stopwords = stopwords

        self.cits = 0
        self.last_print = 0
        self.citations_skipped = 0

        self.jstor_zip_base = Path(jstor_zip_base)
        assert(self.jstor_zip_base.exists())

        self.output_database = output_database
        self.name_blacklist = name_blacklist
        self.RUN_EVERYTHING = RUN_EVERYTHING
        self.groups = groups
        self.group_reps = group_reps
        self.citations_filter = citations_filter
        self.journals_filter = journals_filter
        self.debug = debug

        self.CONSOLIDATE_ITERS = CONSOLIDATE_ITERS
        self.NUM_TERMS_TO_KEEP = NUM_TERMS_TO_KEEP

        # Instantiating counters
        self.ind = defaultdict(lambda: defaultdict(int))
        self.track_doc = defaultdict(lambda: defaultdict(set))
        self.doc = defaultdict(lambda: defaultdict(int))

        self.term_whitelist = term_whitelist





    def cnt(self, term, space, doc):
        if ".".join(sorted(space.split("."))) != space:
            raise Exception(space, "should be sorted...")

        # it's a set, yo
        self.track_doc[space][term].add(doc)
        # update cnt_doc
        self.doc[space][term] = len(self.track_doc[space][term])
        # update ind count
        self.ind[space][term] += 1





    def consolidate_terms(self):
        have_now = set(self.doc['t'])
        # this is where the filtering occurs

        to_keep = set()
        if True:
            # takes terms based on the maximum number I can take...
            terms = list(self.doc['t'].keys())
            counts = np.array([self.doc['t'][k] for k in terms])
            argst = list(reversed(np.argsort(counts)))

            to_keep = [terms[i] for i in argst if '-' in terms[i][0]][
                      :self.NUM_TERMS_TO_KEEP // 2]  # half should be 2-tuples
            to_keep += [terms[i] for i in argst if not '-' in terms[i][0]][
                       :self.NUM_TERMS_TO_KEEP // 2]  # half should be 1-tuples

            to_remove = have_now.difference(to_keep)
            to_remove = set("-".join(x) for x in to_remove)

        if False:
            # takes the top 5000 terms in terms of yearly count
            sort_them = sorted(self.doc['fy.t'], key=lambda x: -self.doc['fy.t'][x])
            to_keep = defaultdict(set)

            i = 0
            while not len(to_keep) or (
                    min(len(x) for x in to_keep.values()) < NPERYEAR and
                    i < len(sort_them)
            ):
                # adds the term to the year set, if it's not already "full"
                me = sort_them[i]
                me_fy, me_t = me

                # eventually, we don't count it :P
                if self.doc['t'][me_t] < CONSOLIDATION_CUTOFF:
                    break

                if len(to_keep[me_fy]) < NPERYEAR:
                    to_keep[me_fy].add(me_t)
                i += 1

            if False:  # useful for debugging
                print({
                    k: len(v)
                    for k, v in to_keep.items()
                })

            to_keep = set(chain.from_iterable(x for x in to_keep.values()))
            to_remove = have_now.difference(to_keep)

        # so that we never log counts for these again:
        self.term_whitelist.update([x[0] for x in to_keep])

        # the rest of the code is pruning all other term counts for this term in memory
        print("consolidating... removing", len(to_remove), 'e.g.', sample(to_remove, 5))

        to_prune = ['t', 'fy.t', 'fj.t', 'c.t']
        for tp in to_prune:

            whichT = tp.split(".").index('t')  # this checks where 't' is in the name of the variable (first or second?)

            print("pruning '%s'..." % tp)

            tydels = [x for x in self.doc[tp] if x[whichT] in to_remove]

            print("old size:", len(self.doc[tp]))
            for tr in tydels:
                del self.doc[tp][tr]
                del self.ind[tp][tr]
            print("new size:", len(self.doc[tp]))

        print("final terms: ", ", ".join(sample(list("-".join(list(x)) for x in self.doc['t']), 200)))


    def account_for(self, doc):

        # consolidating "terms" counter as I go, to limit RAM overhead
        # I'm only interested in the most common 1000
        if self.CONSOLIDATE_ITERS > 0:
            if self.cits - self.last_print > self.CONSOLIDATE_ITERS:
                print("Citation %s" % self.cits)
                print("Term %s" % len(self.doc['t']))
                # print(sample(list(self.doc['t']), 10))
                last_print = self.cits
                self.consolidate_terms()

        if 'citations' not in doc or not len(doc['citations']):
            # print("No citations", doc['doi'])
            return

        for c in doc['citations']:
            if 'contextPure' not in c:
                raise Exception("no contextPure...")

            for cited in c['citations']:

                if self.citations_filter is not None and (cited not in self.citations_filter):
                    citations_skipped += 1
                    continue

                self.cits += 1
                self.cnt(doc['year'], 'fy', doc['doi'])

                # citation
                self.cnt(cited, 'c', doc['doi'])

                # journal
                self.cnt(doc['journal'], 'fj', doc['doi'])

                # journal year
                self.cnt((doc['journal'], doc['year']), 'fj.fy', doc['doi'])

                # citation journal
                self.cnt((cited, doc['journal']), 'c.fj', doc['doi'])

                # citation year
                self.cnt((cited, doc['year']), 'c.fy', doc['doi'])

            # constructing the tuples set :)
            sp = c['contextPure'].lower()
            sp = re.sub("[^a-zA-Z\s]+", "", sp)  # removing extraneous characters
            sp = re.sub("\s+", " ", sp)  # removing extra characters
            sp = sp.strip()
            sp = sp.split()  # splitting into words

            sp = [x for x in sp if x not in self.stopwords]  # strip stopwords

            if False:
                tups = set(zip(sp[:-1], sp[1:]))  # two-word tuples
            elif False:
                tups = set((t1, t2) for t1 in sp for t2 in sp if t1 != t2)  # every two-word pair :)
            else:

                tups = set("-".join(sorted(x)) for x in set(zip(sp[:-1], sp[1:])))  # two-word tuples
                tups.update(sp)  # one-word tuples

            # print(len(tups),c['contextPure'], "---", tups)

            if len(self.term_whitelist):
                tups = [x for x in tups if x in term_whitelist]

            # just term count, in case we are using the `basic` mode
            for t1 in tups:
                # term
                self.cnt((t1,), 't', doc['doi'])

                # term year
                self.cnt((doc['year'], t1), 'fy.t', doc['doi'])

            if self.RUN_EVERYTHING:

                for cited in c['citations']:

                    if use_included_citations_filter and (cited not in included_citations):
                        continue

                    # term features
                    for t1 in tups:

                        # cited work, tuple
                        self.cnt((cited, t1), 'c.t', doc['doi'])

                        # term journal
                        self.cnt((doc['journal'], t1), 'fj.t', doc['doi'])

                        if False:  # eliminating data I'm not using

                            # author loop
                            for a in doc['authors']:
                                # term author
                                self.cnt((a, t1), 'fa.t', doc['doi'])

                        if len(self.term_whitelist):  # really don't want to do this too early. wait until it's narrowed down to the 5k
                            # term term...
                            for t2 in tups:
                                # if they intersect each other, continue...
                                if len(set(t1).intersection(set(t2))) >= min(len(t1), len(t2)):
                                    continue

                                # term term
                                self.cnt((t1, t2), 't.t', doc['doi'])

                    # author loop
                    for a in doc['authors']:
                        # citation author
                        self.cnt((cited, a), 'c.fa', doc['doi'])

                        # year author journal
                        self.cnt((a, doc['journal'], doc['year']), 'fa.fj.fy', doc['doi'])

                        # author
                        self.cnt((a,), 'fa', doc['doi'])

                    # add to counters for citation-citation counts
                    for cited1 in c['citations']:
                        for cited2 in c['citations']:
                            if cited1 >= cited2:
                                continue

                            self.cnt((cited1, cited2), 'c.c', doc['doi'])
                            self.cnt((cited1, cited2, doc['year']), 'c.c.fy', doc['doi'])


    def count(self):

        seen = set()

        skipped = 0

        total_count = Counter()
        doc_count = Counter()
        pair_count = Counter()

        debug = False

        for i, (doi, drep) in enumerate(self.doc_iterator()): #!!!!!!!!!!!!!!!!! fix this now...

            if i % 1000 == 0:
                print("Document", i, "...",
                      len(self.doc['fj'].keys()), "journals...",
                      len(self.doc['c'].keys()), "cited works...",
                      len(self.doc['fa'].keys()), "authors...",
                      len(self.doc['t'].keys()), "terms used...",
                      #citations_skipped, "skipped citations...",
                      self.doc['t'][('social',)], "'social' terms"
                      )

            # sometimes multiple journal names map onto the same journal, for all intents and purposes
            #if drep['journal'] in self.journjournal_map:
            #    drep['journal'] = journal_map[drep['journal']]

                # only include journals in the list "included_journals"
            if self.journals_filter is not None and (drep['journal'] not in self.journals_filter):
                continue

            # now that we have all the information we need,
            # we simply need to "count" this document in a few different ways
            self.account_for(drep)

    def save_counters(self):
        db = Dataset(self.output_database)
        for k, count in self.doc.items():
            varname = "doc ___ %s" % k
            db.save_variable(varname, dict(count))

        for k, count in self.ind.items():
            varname = "ind ___ %s" % k
            db.save_variable(varname, dict(count))


from .. import *

from csv import DictReader
# This cell ensures there are not overflow errors while importing large CSVs

import sys
import csv

maxInt = sys.maxsize

while True:
    # decrease the maxInt value by factor 10
    # as long as the OverflowError occurs.

    try:
        csv.field_size_limit(maxInt)
        break
    except OverflowError:
        maxInt = int(maxInt / 10)


class counter:

    def __init__(
            self,
            wos_txt_base, output_database, name_blacklist=[],
            RUN_EVERYTHING=True,
            groups=None, group_reps=None,
            citations_filter=None, journals_filter=None, debug=False,
            trimCounters=False
    ):

        self.wos_txt_base = Path(wos_txt_base)
        assert (self.wos_txt_base.exists())

        self.name_blacklist = name_blacklist
        self.debug = debug
        self.output_database = output_database
        self.groups = groups
        self.group_reps = group_reps

        self.RUN_EVERYTHING = RUN_EVERYTHING
        self.citations_filter = citations_filter
        self.journals_filter = journals_filter
        self.trimCounters = trimCounters

        # Instantiating counters
        self.ind = defaultdict(lambda: defaultdict(int))
        self.track_doc = defaultdict(lambda: defaultdict(set))
        self.doc = defaultdict(lambda: defaultdict(int))

    def count_cocitations(self):

        allowed_refs = Counter(dict(self.doc['c'].items())).most_common(1000)
        allowed_refs = set(x[0] for x in allowed_refs)

        print("# allowed references for cocitation analysis: %s" % len(allowed_refs))
        print("Examples: %s" % str(list(allowed_refs)[:3]))

        refcount = 0

        for doc in self.doc_iterator():
            refs = list(doc.generate_references())
            for ref1 in refs:
                if ref1.full_ref not in allowed_refs:
                    continue

                for ref2 in refs:
                    if ref2.full_ref not in allowed_refs:
                        continue
                    if ref1.full_ref >= ref2.full_ref:
                        continue

                    self.cnt((ref1.full_ref, ref2.full_ref), 'c1.c2', doc.uid)
                    self.cnt((ref1.publish, ref2.publish), 'c1y.c2y', doc.uid)

                    refcount += 1
                    if refcount % 10000 == 0:
                        print("%s cocitations logged" % refcount)

        print("Finished cocitation logging!")

    def cnt(self, term, space, doc):
        if ".".join(sorted(space.split("."))) != space:
            raise Exception(space, "should be sorted...")

        # it's a set, yo
        self.track_doc[space][term].add(doc)
        # update cnt_doc
        self.doc[space][term] = len(self.track_doc[space][term])
        # update ind count
        self.ind[space][term] += 1

    def doc_iterator(self):
        # processes WoS txt output files one by one, counting relevant cooccurrences as it goes
        self.dcount = 0
        files = list(self.wos_txt_base.glob("**/*.txt"))
        for i, f in enumerate(files):

            with f.open(encoding='utf8', newline='') as pfile:
                r = DictReader(pfile, delimiter='\t', quoting=3)
                rows = list(r)

            if i % 50 == 0:
                print("File %s/%s: %s" % (i, len(files), f.name))
                print("Document: %s" % self.dcount)
                print("Citations: %s" % len(self.doc['c']))

            if i > 10:
                if self.debug:
                    break

            for i, r in enumerate(rows):

                if r['DT'] != "Article":
                    continue

                try:
                    int(r['PY'])
                except ValueError:
                    print(r)
                    raise

                # REMEMBER THIS! journals are in lowercase... not case sensitive

                # implements the journals filter!
                if self.journals_filter is not None and r['SO'].lower() not in self.journals_filter:
                    continue

                yield wos_doc(r)

                self.dcount += 1

    def count_citation(self, doc, ref):

        full_ref = ref.full_ref

        # implement grouping of references
        if self.groups is not None:
            if full_ref in self.groups:
                # retrieves retrospectively-computed groups
                full_ref = self.group_reps[
                    self.groups[full_ref]
                ]
            else:
                raise Exception('not in group...', full_ref)

        ref.full_ref = full_ref

        self.cnt(doc.journal, 'fj', doc.uid)
        self.cnt(doc.publish, 'fy', doc.uid)
        self.cnt(ref.publish, 'ty', doc.uid)
        self.cnt((ref.full_ref, doc.publish), 'c.fy', doc.uid)
        self.cnt((ref.full_ref, doc.journal), 'c.fj', doc.uid)

        self.cnt((doc.journal, doc.publish), 'fj.fy', doc.uid)

        self.cnt(ref.full_ref, 'c', doc.uid)

        if self.RUN_EVERYTHING:

            self.cnt((doc.publish, ref.publish), 'fy.ty', doc.uid)
            self.cnt((doc.journal, ref.publish), 'fj.ty', doc.uid)

            self.cnt(ref.author, 'ta', doc.uid)
            self.cnt((doc.publish, ref.author), 'fy.ta', doc.uid)
            self.cnt((doc.journal, ref.author), 'fj.ta', doc.uid)

            self.cnt((ref.full_ref, doc.journal, doc.publish), 'c.fj.fy', doc.uid)

            # first author!
            ffa = doc.citing_authors[0]
            self.cnt(ffa, 'ffa', doc.uid)
            self.cnt((ffa, doc.publish), 'ffa.fy', doc.uid)
            self.cnt((ffa, doc.journal), 'ffa.fj', doc.uid)
            self.cnt((ref.full_ref, ffa), 'c.ffa', doc.uid)
            # self.cnt((ffa,r['SO'], int(r['PY'])), 'ffa.fj.fy', doc.uid)

            for a in doc.citing_authors:
                self.cnt(a, 'fa', doc.uid)
                self.cnt((a, doc.publish), 'fa.fy', doc.uid)
                self.cnt((a, doc.journal), 'fa.fj', doc.uid)
                # self.cnt((a,r['SO'], int(r['PY'])), 'fa.fj.fy', doc.uid)

                self.cnt((ref.full_ref, a), 'c.fa', doc.uid)

    def main_loop(self):

        for doc in self.doc_iterator():

            for ref in doc.generate_references():

                if ref.author in self.name_blacklist:
                    continue

                if self.groups is not None and ref.full_ref not in self.groups:
                    continue

                self.count_citation(
                    doc,
                    ref
                )

    def save_counters(self):
        db = Dataset(self.output_database)
        db.clear_all()
        for k, count in self.doc.items():
            varname = "doc ___ %s" % k
            db.save_variable(varname, dict(count))

        for k, count in self.ind.items():
            varname = "ind ___ %s" % k
            db.save_variable(varname, dict(count))

    def count_ages(self):

        cbirthdays = defaultdict(lambda: 2050)
        for (c, y), count in self.doc['c.fy'].items():
            if count == 0:
                continue
            cbirthdays[c] = min(cbirthdays[c], y)

        ffabirthdays = defaultdict(lambda: 2050)
        for (c, y), count in self.doc['ffa.fy'].items():
            if count == 0:
                continue
            ffabirthdays[c] = min(ffabirthdays[c], y)

        tabirthdays = defaultdict(lambda: 2050)
        for (y, c), count in self.doc['fy.ta'].items():
            if count == 0:
                continue
            tabirthdays[c] = min(tabirthdays[c], y)

        for doc in self.doc_iterator():

            refs = list(doc.generate_references())
            for i, ref in enumerate(refs):

                if ref.author in self.name_blacklist:
                    continue

                if self.groups is not None and ref.full_ref not in self.groups:
                    continue

                c = ref.full_ref
                ffa = doc.citing_authors[0]
                ta = ref.author

                # skips the hitherto uncounted
                if c not in self.doc['c'] or self.doc['c'][c] == 0:
                    continue
                if ffa not in self.doc['ffa'] or self.doc['ffa'][ffa] == 0:
                    continue
                if ta not in self.doc['ta'] or self.doc['ta'][ta] == 0:
                    continue

                cage1 = doc.publish - cbirthdays[c]
                ffaage1 = doc.publish - ffabirthdays[ffa]
                taage1 = doc.publish - tabirthdays[ta]

                if not all(x >= 0 for x in [cage1, ffaage1, taage1]):
                    print(cage1, ffaage1, taage1)
                    print(self.doc['c'][ref.full_ref], self.doc['ffa'][doc.citing_authors[0]],
                          self.doc['ta'][ref.author])
                    print(ref.full_ref, doc.citing_authors[0], ref.author)
                    print(cage1, ffaage1, taage1)
                    raise

                self.cnt((cage1, doc.journal), 'cAge.fj', doc.uid)

                self.cnt((cage1, doc.citing_authors[0]), 'cAge.ffa', doc.uid)
                for author in doc.citing_authors:
                    faage1 = doc.publish - ffabirthdays[author]
                    self.cnt((cage1, author), 'cAge.fa', doc.uid)
                    self.cnt(faage1, 'faAge', doc.uid)

                    self.cnt((faage1, ref.author), 'faAge.ta', doc.uid)
                    self.cnt((faage1, doc.publish, ref.author), 'faAge.fy.ta', doc.uid)
                    self.cnt((faage1, taage1), 'faAge.taAge', doc.uid)
                    self.cnt((cage1, faage1), 'cAge.faAge', doc.uid)

                    self.cnt((faage1, doc.publish), 'faAge.fy', doc.uid)

                self.cnt(ffaage1, 'ffaAge', doc.uid)

                self.cnt((ffaage1, ref.author), 'ffaAge.ta', doc.uid)
                self.cnt((ffaage1, doc.publish, ref.author), 'ffaAge.fy.ta', doc.uid)
                self.cnt((ffaage1, taage1), 'ffaAge.taAge', doc.uid)
                self.cnt((cage1, ffaage1), 'cAge.ffaAge', doc.uid)

                self.cnt((cage1, doc.publish), 'cAge.fy', doc.uid)
                self.cnt((doc.publish, taage1), 'fy.taAge', doc.uid)
                self.cnt((ffaage1, doc.publish), 'ffaAge.fy', doc.uid)

                for j, ref2 in enumerate(refs):
                    cage2 = doc.publish - cbirthdays[ref2.full_ref]
                    if i >= j:
                        continue

                    self.cnt((cage1, cage2), 'c1age.c2age', doc.uid)

    def run(self):
        self.main_loop()
        if self.RUN_EVERYTHING:
            self.count_ages()
            self.count_cocitations()
        self.save_counters()


from .env import GLOBS, cfile

if __name__ == '__main__':
    import sys
    from pathlib import Path

    args = sys.argv[1:]

    me = Path(__file__).parent

    if not len(args):
        raise Exception("You need to specify a command: `init`, `clone`, `start`")

    if args[0] == 'init':

        import yaml
        from os.path import expanduser
        home = expanduser("~")
        default = Path(home).joinpath("knowknow")

        if 'kkdir' not in GLOBS:
            while 1:
                chosen_dir = input("Enter the directory where knowknow will keep data and code (will be created if doesn't exist) <default: %s> : " % default) or str(default)
                chosen_dir = Path(chosen_dir)
                if not chosen_dir.parent.exists():
                    ans = None
                    while ans not in "ynYN":
                        ans = input("The parent of this directory does not exist. Continue (y/n)?")
                    if ans in "nN":
                        continue

                GLOBS['kkdir'] = str(chosen_dir)

                with cfile.open('w') as f:
                    yaml.dump(GLOBS, f)

                if not chosen_dir.exists():
                    print("Creating directory.")
                    chosen_dir.mkdir(parents=True)
                else:
                    #print("This directory exists.")
                    pass

                print("Dataset directory updated")
                break

        dir_create = Path(GLOBS['kkdir']).joinpath('code',args[1])
        print("Creating directory '%s'" % dir_create)
        dir_create.mkdir(exist_ok=True, parents=True)

    elif args[0] == 'clone':

        if len(args) != 2:
            raise Exception('`clone` command takes exactly 1 argument')


        if 'kkdir' not in GLOBS:
            raise Exception('You need to call `python -m knowknow init` first...')

        _, name = args

        if name[0] == '@':
            name = name[1:]

        if not( 'github.com' in name ):
            name = "https://github.com/%s" % name

        short_name = name.split("/")[-1]

        print("Cloning '%s' into '<kk>/code/%s' ..." % (name, short_name))

        from git.repo.base import Repo
        fld = Path(GLOBS['kkdir'], 'code')
        if not fld.exists():
            fld.mkdir()
        Repo.clone_from(name, fld.joinpath(short_name))

    elif args[0] == 'start':
        if len(args) != 2:
            fs = list(Path(GLOBS['kkdir'], 'code').glob("*"))
            if not len(fs):
                poss_args = "No code has been loaded, so there aren't valid commands. Either run `init` or `clone` to create/copy some data."
            else:
                poss_args = "Currently loaded: %s" %  ', '.join("`%s`"%x.name for x in fs)
            raise Exception('`start` command takes exactly 1 argument... %s' % poss_args)

        where = args[1]
        where = where.split("/")[-1]

        dr = Path(GLOBS['kkdir'], 'code', where)

        import os
        os.chdir(dr)
        os.environ['NBDIR'] = str(dr)
        os.system('jupyter-lab')

    elif args[0] == 'help':
        print("Possible commands are clone (with GitHub url), init (no args), and start (no args)")

    else:
        print("Command not recognized.")
        print("Possible commands are `clone` (with GitHub url), `init` (no args), and `start` (code directory)")